/*\
|*| --------------------------------------------------------------------------
|*|	S3 SDAC clock synthesizer/ramdac driver - ramdac part 
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995, The GGI Project.
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"

#ifdef DEBUG
  #define	NEED_GRAPHTYPE_SYMS
#endif

#include <ggi/graphdev.h>
#include "S3/sdac.h"

#define	DAC_PRECISION	6	/* we can do 6 bit R,G,B resolution	*/
#include "IBM/clut.inc"

#define DAC_VERSION "0.00 ALPHA"

#if !((MONITORVERSION == 01) && (MODEVERSION == 03) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif

#define	SUPPORTED (GM_TEXT | GM_1BIT | GM_4BIT | GM_8BIT \
		  | GM_15BIT | GM_16BIT | GM_32BIT | GM_4BIT_PLANES)
#define	CLUT_USED (GM_TEXT | GM_1BIT | GM_4BIT | GM_4BIT_PLANES | GM_8BIT)

static struct kgi_ramdac ramdac =
       { "S3 Incorporated",		/* manufacturer		*/
	 "86C716-ME/MG SDAC",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 0,				/* flags		*/
	 { 0x7FFF, 0x7FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_16BIT_PB,
	 {        0, 110000000 },	/* normal mode		*/
	 { 67500000, 135000000 },	/* 2xclk mode		*/

	 {{ 	   0, 110000000 },	/* 	GT_TEXT		*/
	 {	   0,         0 },	/* 	GT_1BIT		*/
	 {	   0, 110000000 },	/* 	GT_4BIT		*/
	 {	   0, 110000000 },	/* 	GT_8BIT		*/	
	 {	   0, 110000000 },	/* 	GT_15BIT	*/
	 { 	   0, 110000000 },	/* 	GT_16BIT	*/
	 {	   0,  73000000 },	/* 	GT_24BIT	*/
	 {	   0,  55000000 },	/* 	GT_32BIT	*/
	 {	   0, 110000000 }}	/* 	GT_4BIT_PLANES	*/
	};

/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
  int id, old;

  old = inbDAC(DAC_PIXEL_MASK);
  outbDAC(0xFF, DAC_PIXEL_MASK);
  inbDAC(DAC_PW_INDEX);
  inbDAC(DAC_PIXEL_MASK);
  inbDAC(DAC_PIXEL_MASK);
  inbDAC(DAC_PIXEL_MASK);
  id = inbDAC(DAC_PIXEL_MASK);
  outbDAC(old, DAC_PIXEL_MASK);

  if ((id & DAC_ID_MASK) == DAC_SDAC_ID)
    {
      TRACE(printk("  %s %s rev %i ramdac driver.\n", ramdac.manufact,
                   ramdac.model, id & DAC_REV_MASK));
      return EOK;
    }
  printk(" %s %s not detected.\n", ramdac.manufact, ramdac.model);
  return -ERAMDAC;
}

void
kgi_ramdac_done(void)
{
  TRACE(printk("%s %s driver removed.\n", ramdac.manufact, ramdac.model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return &ramdac; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int mul = mode->lclk.mul, div = mode->lclk.div;
  int newclock = mode->dclk;

  switch (mode->request.graphtype)
   {				/* first try a setup usign PA[15:0]	*/
     case GT_TEXT:	mul = 1; div = 1; break;

     case GT_1BIT:
     case GT_4BIT:
     case GT_4BIT_PLANES:
     case GT_8BIT:	mul = 1; div = 2; break;
	
     case GT_15BIT:
     case GT_16BIT:	mul = 1; div = 1; break;

     case GT_24BIT:	mul = 3; div = 1; break;

     case GT_32BIT:	mul = 2; div = 1; break;

     default:		return -ERAMDAC;
   }
  
  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mul; mode->clk.mul = 1;
	mode->lclk.div = div; mode->clk.div = 1;
	TRACE(printk("%s:%i: proposing %i:%i multiplexing mode\n", \
	      __FILE__,__LINE__, mul, div));
	return EOK;

     case CMD_LOWER:
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  return -ERAMDAC;
	
	if (ramdac.clock[mode->request.graphtype].max < newclock)
	  newclock = ramdac.clock[mode->request.graphtype].max;

	if (ramdac.normal.max < newclock*mul/div)
	  newclock = ramdac.normal.max*div/mul;

	mode->dclk = newclock;
	TRACE(printk("%s:%i: %i:%i multiplex mode, DCLK = %i Hz\n", \
	      __FILE__,__LINE__, mul, div, mode->dclk));
	return EOK;

     case CMD_RAISE:
	if (   (ramdac.clock[mode->request.graphtype].max < newclock)
	    || (ramdac.normal.max < newclock))
	  return -ERAMDAC;

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	mode->dclk = newclock;
	TRACE(printk("%s:%i: %i:%i multiplex mode, DCLK = %i Hz\n", \
	      __FILE__,__LINE__, mul, div, mode->dclk));
	return EOK;

     case CMD_CHECK:
/* !!! implement checking! */
	#warning mode_check not implemented yet!
	return EOK;

    default:
	INTERNALERROR;
	return -ERAMDAC;
  }
}

void
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  int ECR;

  switch(mode->request.graphtype)
   {
     case GT_TEXT:		/* all these modes are assumed to fill up the */
     case GT_1BIT:		/* missing bits to one byte. However, these   */
     case GT_4BIT:		/* may be transmitted 8 or 16 bit on the bus. */
     case GT_4BIT_PLANES:
     case GT_8BIT:
	if (mode->lclk.div == 1)
	  ECR = DAC_C_MODE0;	/* transferred via PA[7:0]	*/
         else
          ECR = DAC_C_MODE8;	/* transferred via PA[15:0]	*/
	break;
	
     case GT_15BIT:
	if (mode->lclk.mul == 2)
	  ECR = DAC_C_MODE1;	/* also MODE4 or MODE5 possible	*/
	 else
	  ECR = DAC_C_MODE9;	/* transferred via PA[15:0]	*/
	break;

     case GT_16BIT:
	if (mode->lclk.mul == 2)
	  ECR = DAC_C_MODE3;	/* also MODE6 possible		*/
	 else
	  ECR = DAC_C_MODEA;	/* transferred via PA[15:0]	*/
	break;

     case GT_24BIT:
	if (mode->lclk.div ==1)
	  ECR = DAC_C_MODE2;	/* also MODE7 possible		*/ 
	 else
	  ECR = DAC_C_MODEC;	/* packed 24bpp via PA[15:0]	*/
	break;

     case GT_32BIT:
	ECR = DAC_C_MODEB;
	break;

     default:
	ECR = 0;
	INTERNALERROR;		/* for safety, and to trick gcc */
   }
  TRACE(printk("%s: graphtype = %s, ECR = 0x%X\n", \
        __FILE__, GRAPHMASK(mode->request.graphtype), ECR));

  outbDAC(ECR, DAC_COMMAND);

  if ((1 << mode->request.graphtype) & CLUT_USED)
    set_CLUT_entries(mode->palette);
}

int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  /* this may set
       -  dac_params
       -  CLUT (or gamma correction table).
   */

  if (get_CLUT_entries == NULL)	/* !!! trick gcc	*/
    return -ERAMDAC;
    
  return -ERAMDAC;
}
