/*\
|*| --------------------------------------------------------------------------
|*|     GGI graphics card detection using BIOS strings
|*| --------------------------------------------------------------------------
|*|
|*|     Copyright (C) 1996 by
|*|
|*|       Andreas Beck,         [becka@sunserver1.rz.uni-duesseldorf.de]
|*|
|*|     This program is free software; you can redistribute it and/or modify
|*|     it under the terms of the GNU General Public License as published by
|*|     the Free Software Foundation; either version 2, or (at your option)
|*|     any later version.
|*|
|*|     You should have received a copy of the GNU General Public License
|*|     along with this program; see the file COPYING.  If not, write to
|*|     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define MAINTAINER      "Andreas Beck, [becka@sunserver1.rz.uni-duesseldorf.de]"

#include <stdio.h>
#include <stdlib.h>

/*\ This program expects to be given a database as the first argument.
|*| Format:
|*| # comment
|*| POS,ANZ,"ID-String"
|*| Cardname
|*| CHIP_CLASS,CHIP_TYPE
|*| CLK_CLASS ,CLK_TYPE
|*| DAC_CLASS ,DAC_TYPE
\*/

void unesc(char *ptr)
{ char input[256],*ph;
  int h;
  strcpy(input,ptr);ph=input+1;
  do {
    switch(*ph)
    { case '"':*ptr++='\0';return;	/* terminator */
      case '\\':switch(*(++ph))
               { case '\\': *ptr++=*ph++;break;
                 case 'x': ph++;sscanf(ph,"%2x",&h);*ptr++=(char)h;ph+=2;break;
                 default : fprintf(stderr,"unknown escape sequence %s.\n",ph);exit(EXIT_FAILURE);
                 	   break;
               } break;
      default :*ptr++=*ph++;break;
    }
  } while(1);
}

int main(int argc,char *argv[])
{ FILE *db;
  char bios[32*1024];
  char help[256],idstr[256];
  
  int pos,anz,line;
  
  line=0;
  if (argc!=2) 
    { fprintf(stderr,"usage: biosdetect database\n");return EXIT_FAILURE; }
  if (fread(bios,1,sizeof(bios),stdin)!=sizeof(bios))
    { fprintf(stderr,"cannot read bios from stdin.\n");return EXIT_FAILURE; }
  if ((db=fopen(argv[1],"r"))==NULL) 
    { fprintf(stderr,"cannot open %s.\n",argv[1]);return EXIT_FAILURE; }
  do {
    line++;
    if (fscanf(db,"%[^\n]\n",help)!=1) break;
    if (*help=='#'||*help==';'||*help=='\0'||*help=='\n') continue;
    
    if (sscanf(help,"%d,%d,%[^\n]\n",&pos,&anz,idstr)!=3)
      { fprintf(stderr,"parse error (no ID description) at line %d.\n",line);return EXIT_FAILURE; }
    unesc(idstr);
    if (memcmp(bios+pos,idstr,anz)!=0)
    { if (fscanf(db,"%[^\n]\n",help)!=1) break;
      fprintf(stderr,"Trying: %-60s\r",help);
      for(anz=0;anz<3;anz++)
        if (fscanf(db,"%[^\n]\n",help)!=1) break;
    } else
    { if (fscanf(db,"%[^\n]\n",help)!=1) break;
      printf("# automatic detection : %s\n",help);
      
      if (fscanf(db,"%[^,],%[^\n]\n",idstr,help)!=2) break;
      printf("CHIP_CLASS=%s\nCHIP_TYPE=%s\n",idstr,help);

      if (fscanf(db,"%[^,],%[^\n]\n",idstr,help)!=2) break;
      printf("CLK_CLASS=%s\nCLK_TYPE=%s\n",idstr,help);

      if (fscanf(db,"%[^,],%[^\n]\n",idstr,help)!=2) break;
      printf("DAC_CLASS=%s\nDAC_TYPE=%s\n",idstr,help);
      fclose(db);
      return EXIT_SUCCESS;
    }
  }
  while(1);
  fclose(db);
  return EXIT_SUCCESS;
}
